#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Airflow System Test DAG that verifies Datastore commit operators.
"""

from __future__ import annotations

import os
from datetime import datetime
from typing import Any

from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.datastore import (
    CloudDatastoreAllocateIdsOperator,
    CloudDatastoreBeginTransactionOperator,
    CloudDatastoreCommitOperator,
    CloudDatastoreDeleteOperationOperator,
    CloudDatastoreExportEntitiesOperator,
    CloudDatastoreGetOperationOperator,
    CloudDatastoreImportEntitiesOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "datastore_commit"
BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"
# [START how_to_keys_def]
KEYS = [
    {
        "partitionId": {"projectId": PROJECT_ID, "namespaceId": ""},
        "path": {"kind": "airflow"},
    }
]
# [END how_to_keys_def]

# [START how_to_transaction_def]
TRANSACTION_OPTIONS: dict[str, Any] = {"readWrite": {}}
# [END how_to_transaction_def]


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "datastore"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket", bucket_name=BUCKET_NAME, project_id=PROJECT_ID, location="EU"
    )

    # [START how_to_allocate_ids]
    allocate_ids = CloudDatastoreAllocateIdsOperator(
        task_id="allocate_ids", partial_keys=KEYS, project_id=PROJECT_ID
    )
    # [END how_to_allocate_ids]

    # [START how_to_begin_transaction]
    begin_transaction_commit = CloudDatastoreBeginTransactionOperator(
        task_id="begin_transaction_commit",
        transaction_options=TRANSACTION_OPTIONS,
        project_id=PROJECT_ID,
    )
    # [END how_to_begin_transaction]

    # [START how_to_commit_def]
    COMMIT_BODY = {
        "mode": "TRANSACTIONAL",
        "mutations": [
            {
                "insert": {
                    "key": KEYS[0],
                    "properties": {"string": {"stringValue": "airflow is awesome!"}},
                }
            }
        ],
        "singleUseTransaction": {"readWrite": {}},
    }
    # [END how_to_commit_def]

    # [START how_to_commit_task]
    commit_task = CloudDatastoreCommitOperator(task_id="commit_task", body=COMMIT_BODY, project_id=PROJECT_ID)
    # [END how_to_commit_task]

    # [START how_to_export_task]
    export_task = CloudDatastoreExportEntitiesOperator(
        task_id="export_task",
        bucket=BUCKET_NAME,
        project_id=PROJECT_ID,
        overwrite_existing=True,
    )
    # [END how_to_export_task]

    # [START how_to_import_task]
    import_task = CloudDatastoreImportEntitiesOperator(
        task_id="import_task",
        bucket="{{ task_instance.xcom_pull('export_task')['response']['outputUrl'].split('/')[2] }}",
        file="{{ '/'.join(task_instance.xcom_pull('export_task')['response']['outputUrl'].split('/')[3:]) }}",
        project_id=PROJECT_ID,
    )
    # [END how_to_import_task]

    # [START get_operation_state]
    get_operation = CloudDatastoreGetOperationOperator(
        task_id="get_operation", name="{{ task_instance.xcom_pull('export_task')['name'] }}"
    )
    # [END get_operation_state]

    # [START delete_operation]
    delete_export_operation = CloudDatastoreDeleteOperationOperator(
        task_id="delete_export_operation",
        name="{{ task_instance.xcom_pull('export_task')['name'] }}",
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END delete_operation]

    delete_import_operation = CloudDatastoreDeleteOperationOperator(
        task_id="delete_import_operation",
        name="{{ task_instance.xcom_pull('import_task')['name'] }}",
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    chain(
        # TEST SETUP
        create_bucket,
        # TEST BODY
        allocate_ids,
        begin_transaction_commit,
        commit_task,
        export_task,
        import_task,
        get_operation,
        # TEST TEARDOWN
        [delete_bucket, delete_export_operation, delete_import_operation],
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.google.cloud.utils.helpers import normalize_directory_path, resource_path_to_dict


class TestHelpers:
    def test_normalize_directory_path(self):
        assert normalize_directory_path("dir_path") == "dir_path/"
        assert normalize_directory_path("dir_path/") == "dir_path/"
        assert normalize_directory_path(None) is None

    def test_resource_path_to_dict(self):
        resource_name = "key1/value1/key2/value2"
        expected_dict = {"key1": "value1", "key2": "value2"}
        actual_dict = resource_path_to_dict(resource_name=resource_name)
        assert set(actual_dict.items()) == set(expected_dict.items())

    def test_resource_path_to_dict_empty(self):
        resource_name = ""
        expected_dict = {}
        assert resource_path_to_dict(resource_name=resource_name) == expected_dict

    def test_resource_path_to_dict_fail(self):
        with pytest.raises(ValueError):
            resource_path_to_dict(resource_name="key/value/key")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import base64
import binascii
import sys
from unittest import mock

import dill
import pytest

if sys.version_info < (3, 12):
    from airflow.providers.google.cloud.utils import mlengine_prediction_summary
else:
    pytestmark = pytest.mark.skip(
        f"package apache_beam is not available for Python 3.12. Skipping all tests in {__name__}"
    )


class TestJsonCode:
    def test_encode(self):
        assert mlengine_prediction_summary.JsonCoder.encode({"a": 1}) == b'{"a": 1}'

    def test_decode(self):
        assert mlengine_prediction_summary.JsonCoder.decode('{"a": 1}') == {"a": 1}


class TestMakeSummary:
    def test_make_summary(self):
        print(mlengine_prediction_summary.MakeSummary(1, lambda x: x, []))

    def test_run_without_all_arguments_should_raise_exception(self):
        with pytest.raises(SystemExit):
            mlengine_prediction_summary.run()

        with pytest.raises(SystemExit):
            mlengine_prediction_summary.run(
                [
                    "--prediction_path=some/path",
                ]
            )

        with pytest.raises(SystemExit):
            mlengine_prediction_summary.run(
                [
                    "--prediction_path=some/path",
                    "--metric_fn_encoded=encoded_text",
                ]
            )

    def test_run_should_fail_for_invalid_encoded_fn(self):
        with pytest.raises(binascii.Error):
            mlengine_prediction_summary.run(
                [
                    "--prediction_path=some/path",
                    "--metric_fn_encoded=invalid_encoded_text",
                    "--metric_keys=a",
                ]
            )

    def test_run_should_fail_if_enc_fn_is_not_callable(self):
        non_callable_value = 1
        fn_enc = base64.b64encode(dill.dumps(non_callable_value)).decode("utf-8")

        with pytest.raises(ValueError):
            mlengine_prediction_summary.run(
                [
                    "--prediction_path=some/path",
                    "--metric_fn_encoded=" + fn_enc,
                    "--metric_keys=a",
                ]
            )

    def test_run_should_not_fail_with_valid_fn(self):
        with (
            mock.patch.object(mlengine_prediction_summary.beam.pipeline, "PipelineOptions") as pipeline_mock,
            mock.patch.object(mlengine_prediction_summary.beam, "Pipeline") as pipeline_obj_mock,
            mock.patch.object(mlengine_prediction_summary.beam.io, "ReadFromText") as io_mock,
        ):

            def metric_function():
                return 1

            fn_enc = base64.b64encode(dill.dumps(metric_function)).decode("utf-8")

            mlengine_prediction_summary.run(
                [
                    "--prediction_path=some/path",
                    "--metric_fn_encoded=" + fn_enc,
                    "--metric_keys=a",
                ]
            )

            pipeline_mock.assert_called_once_with([])
            pipeline_obj_mock.assert_called_once()
            io_mock.assert_called_once()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG for Dataproc batch operators.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.api_core.retry import Retry

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.dataproc import (
    ClusterGenerator,
    DataprocCreateBatchOperator,
    DataprocCreateClusterOperator,
    DataprocDeleteBatchOperator,
    DataprocDeleteClusterOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "dataproc_batch_ps"
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}".replace("-", "_")
REGION = "us-east4"
CLUSTER_NAME_BASE = f"cluster-{DAG_ID}".replace("_", "-")
CLUSTER_NAME_FULL = CLUSTER_NAME_BASE + f"-{ENV_ID}".replace("_", "-")
CLUSTER_NAME = CLUSTER_NAME_BASE if len(CLUSTER_NAME_FULL) >= 33 else CLUSTER_NAME_FULL
BATCH_ID = f"batch-{ENV_ID}-{DAG_ID}".replace("_", "-")

CLUSTER_GENERATOR_CONFIG_FOR_PHS = ClusterGenerator(
    project_id=PROJECT_ID,
    region=REGION,
    master_machine_type="n1-standard-4",
    worker_machine_type="n1-standard-4",
    num_workers=0,
    properties={
        "spark:spark.history.fs.logDirectory": f"gs://{BUCKET_NAME}",
    },
    enable_component_gateway=True,
).make()
BATCH_CONFIG_WITH_PHS = {
    "spark_batch": {
        "jar_file_uris": ["file:///usr/lib/spark/examples/jars/spark-examples.jar"],
        "main_class": "org.apache.spark.examples.SparkPi",
    },
    "environment_config": {
        "peripherals_config": {
            "spark_history_server_config": {
                "dataproc_cluster": f"projects/{PROJECT_ID}/regions/{REGION}/clusters/{CLUSTER_NAME}"
            }
        }
    },
}


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "dataproc", "batch", "persistent"],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket", bucket_name=BUCKET_NAME, project_id=PROJECT_ID
    )

    # [START how_to_cloud_dataproc_create_cluster_for_persistent_history_server]
    create_cluster = DataprocCreateClusterOperator(
        task_id="create_cluster_for_phs",
        project_id=PROJECT_ID,
        cluster_config=CLUSTER_GENERATOR_CONFIG_FOR_PHS,
        region=REGION,
        cluster_name=CLUSTER_NAME,
        retry=Retry(maximum=100.0, initial=10.0, multiplier=1.0),
        num_retries_if_resource_is_not_ready=3,
    )
    # [END how_to_cloud_dataproc_create_cluster_for_persistent_history_server]

    # [START how_to_cloud_dataproc_create_batch_operator_with_persistent_history_server]
    create_batch = DataprocCreateBatchOperator(
        task_id="create_batch_with_phs",
        project_id=PROJECT_ID,
        region=REGION,
        batch=BATCH_CONFIG_WITH_PHS,
        batch_id=BATCH_ID,
        result_retry=Retry(maximum=100.0, initial=10.0, multiplier=1.0),
        num_retries_if_resource_is_not_ready=3,
    )
    # [END how_to_cloud_dataproc_create_batch_operator_with_persistent_history_server]

    delete_batch = DataprocDeleteBatchOperator(
        task_id="delete_batch",
        project_id=PROJECT_ID,
        region=REGION,
        batch_id=BATCH_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_cluster = DataprocDeleteClusterOperator(
        task_id="delete_cluster",
        project_id=PROJECT_ID,
        cluster_name=CLUSTER_NAME,
        region=REGION,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    (
        # TEST SETUP
        create_bucket
        >> create_cluster
        # TEST BODY
        >> create_batch
        # TEST TEARDOWN
        >> delete_batch
        >> delete_cluster
        >> delete_bucket
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import asyncio
from datetime import datetime, timedelta, timezone
from typing import Any
from unittest import mock
from unittest.mock import AsyncMock

import pytest
from gcloud.aio.storage import Bucket, Storage

from airflow.providers.google.cloud.hooks.gcs import GCSAsyncHook
from airflow.providers.google.cloud.triggers.gcs import (
    GCSBlobTrigger,
    GCSCheckBlobUpdateTimeTrigger,
    GCSPrefixBlobTrigger,
    GCSUploadSessionTrigger,
)
from airflow.triggers.base import TriggerEvent

TEST_BUCKET = "TEST_BUCKET"
TEST_OBJECT = "TEST_OBJECT"
TEST_PREFIX = "TEST_PREFIX"
TEST_GCP_CONN_ID = "TEST_GCP_CONN_ID"
TEST_POLLING_INTERVAL = 3.0
TEST_HOOK_PARAMS: dict[str, Any] = {}
TEST_TS_OBJECT = datetime.now(tz=timezone.utc)


TEST_INACTIVITY_PERIOD = 5.0
TEST_MIN_OBJECTS = 1
TEST_ALLOW_DELETE = True
TEST_PREVIOUS_OBJECTS = {"a", "ab"}


@pytest.fixture
def trigger():
    return GCSBlobTrigger(
        bucket=TEST_BUCKET,
        object_name=TEST_OBJECT,
        use_glob=False,
        poke_interval=TEST_POLLING_INTERVAL,
        google_cloud_conn_id=TEST_GCP_CONN_ID,
        hook_params=TEST_HOOK_PARAMS,
    )


@pytest.fixture
def trigger_using_glob():
    return GCSBlobTrigger(
        bucket=TEST_BUCKET,
        object_name=TEST_OBJECT,
        use_glob=True,
        poke_interval=TEST_POLLING_INTERVAL,
        google_cloud_conn_id=TEST_GCP_CONN_ID,
        hook_params=TEST_HOOK_PARAMS,
    )


class TestGCSBlobTrigger:
    def test_gcs_blob_trigger_serialization(self, trigger):
        """
        Asserts that the GCSBlobTrigger correctly serializes its arguments
        and classpath.
        """

        classpath, kwargs = trigger.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.gcs.GCSBlobTrigger"
        assert kwargs == {
            "bucket": TEST_BUCKET,
            "object_name": TEST_OBJECT,
            "use_glob": False,
            "poke_interval": TEST_POLLING_INTERVAL,
            "google_cloud_conn_id": TEST_GCP_CONN_ID,
            "hook_params": TEST_HOOK_PARAMS,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSBlobTrigger._object_exists")
    async def test_gcs_blob_trigger_success(self, mock_object_exists, trigger):
        """
        Tests that the GCSBlobTrigger is success case
        """
        mock_object_exists.return_value = "success"

        generator = trigger.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "success", "message": "success"}) == actual

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSBlobTrigger._object_exists")
    async def test_gcs_blob_trigger_pending(self, mock_object_exists, trigger):
        """
        Test that GCSBlobTrigger is in loop if file isn't found.
        """
        mock_object_exists.return_value = "pending"

        task = asyncio.create_task(trigger.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSBlobTrigger._object_exists")
    async def test_gcs_blob_trigger_exception(self, mock_object_exists, trigger):
        """
        Tests the GCSBlobTrigger does fire if there is an exception.
        """
        mock_object_exists.side_effect = AsyncMock(side_effect=Exception("Test exception"))

        task = [i async for i in trigger.run()]
        assert len(task) == 1
        assert TriggerEvent({"status": "error", "message": "Test exception"}) in task

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "exists,response",
        [
            (True, "success"),
            (False, "pending"),
        ],
    )
    async def test_object_exists(self, exists, response, trigger):
        """
        Tests to check if a particular object in Google Cloud Storage
        is found or not
        """
        hook = AsyncMock(GCSAsyncHook)
        storage = AsyncMock(Storage)
        hook.get_storage_client.return_value = storage
        bucket = AsyncMock(Bucket)
        storage.get_bucket.return_value = bucket
        bucket.blob_exists.return_value = exists

        res = await trigger._object_exists(hook, TEST_BUCKET, TEST_OBJECT)
        assert res == response
        bucket.blob_exists.assert_called_once_with(blob_name=TEST_OBJECT)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "blob_list,response",
        [
            ([TEST_OBJECT], "success"),
            ([], "pending"),
        ],
    )
    async def test_object_exists_using_glob(self, blob_list, response, trigger_using_glob):
        """
        Tests to check if a particular object in Google Cloud Storage
        is found or not
        """
        hook = AsyncMock(GCSAsyncHook)
        storage = AsyncMock(Storage)
        hook.get_storage_client.return_value = storage
        bucket = AsyncMock(Bucket)
        storage.get_bucket.return_value = bucket
        bucket.list_blobs.return_value = blob_list

        res = await trigger_using_glob._object_exists(hook, TEST_BUCKET, TEST_OBJECT)
        assert res == response
        bucket.list_blobs.assert_called_once_with(match_glob=TEST_OBJECT)


class TestGCSPrefixBlobTrigger:
    TRIGGER = GCSPrefixBlobTrigger(
        bucket=TEST_BUCKET,
        prefix=TEST_PREFIX,
        poke_interval=TEST_POLLING_INTERVAL,
        google_cloud_conn_id=TEST_GCP_CONN_ID,
        hook_params=TEST_HOOK_PARAMS,
    )

    def test_gcs_prefix_blob_trigger_serialization(self):
        """
        Asserts that the GCSPrefixBlobTrigger correctly serializes its arguments
        and classpath.
        """

        classpath, kwargs = self.TRIGGER.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.gcs.GCSPrefixBlobTrigger"
        assert kwargs == {
            "bucket": TEST_BUCKET,
            "prefix": TEST_PREFIX,
            "poke_interval": TEST_POLLING_INTERVAL,
            "google_cloud_conn_id": TEST_GCP_CONN_ID,
            "hook_params": TEST_HOOK_PARAMS,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSPrefixBlobTrigger._list_blobs_with_prefix")
    async def test_gcs_prefix_blob_trigger_success(self, mock_list_blobs_with_prefixs):
        """
        Tests that the GCSPrefixBlobTrigger is success case
        """
        mock_list_blobs_with_prefixs.return_value = ["success"]

        generator = self.TRIGGER.run()
        actual = await generator.asend(None)
        assert (
            TriggerEvent({"status": "success", "message": "Successfully completed", "matches": ["success"]})
            == actual
        )

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSPrefixBlobTrigger._list_blobs_with_prefix")
    async def test_gcs_prefix_blob_trigger_exception(self, mock_list_blobs_with_prefixs):
        """
        Tests the GCSPrefixBlobTrigger does fire if there is an exception.
        """
        mock_list_blobs_with_prefixs.side_effect = AsyncMock(side_effect=Exception("Test exception"))

        task = [i async for i in self.TRIGGER.run()]
        assert len(task) == 1
        assert TriggerEvent({"status": "error", "message": "Test exception"}) in task

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSPrefixBlobTrigger._list_blobs_with_prefix")
    async def test_gcs_prefix_blob_trigger_pending(self, mock_list_blobs_with_prefixs):
        """
        Test that GCSPrefixBlobTrigger is in loop if file isn't found.
        """
        mock_list_blobs_with_prefixs.return_value = []

        task = asyncio.create_task(self.TRIGGER.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    async def test_list_blobs_with_prefix(self):
        """
        Tests to check if a particular object in Google Cloud Storage
        is found or not
        """
        hook = AsyncMock(GCSAsyncHook)
        storage = AsyncMock(Storage)
        hook.get_storage_client.return_value = storage
        bucket = AsyncMock(Bucket)
        storage.get_bucket.return_value = bucket
        bucket.list_blobs.return_value = ["test_string"]

        res = await self.TRIGGER._list_blobs_with_prefix(hook, TEST_BUCKET, TEST_PREFIX)
        assert res == ["test_string"]
        bucket.list_blobs.assert_called_once_with(prefix=TEST_PREFIX)


class TestGCSCheckBlobUpdateTimeTrigger:
    TRIGGER = GCSCheckBlobUpdateTimeTrigger(
        bucket=TEST_BUCKET,
        object_name=TEST_OBJECT,
        target_date=TEST_TS_OBJECT,
        poke_interval=TEST_POLLING_INTERVAL,
        google_cloud_conn_id=TEST_GCP_CONN_ID,
        hook_params=TEST_HOOK_PARAMS,
    )

    def test_gcs_blob_update_trigger_serialization(self):
        """
        Asserts that the GCSCheckBlobUpdateTimeTrigger correctly serializes its arguments
        and classpath.
        """

        classpath, kwargs = self.TRIGGER.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.gcs.GCSCheckBlobUpdateTimeTrigger"
        assert kwargs == {
            "bucket": TEST_BUCKET,
            "object_name": TEST_OBJECT,
            "target_date": TEST_TS_OBJECT,
            "poke_interval": TEST_POLLING_INTERVAL,
            "google_cloud_conn_id": TEST_GCP_CONN_ID,
            "hook_params": TEST_HOOK_PARAMS,
        }

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.cloud.triggers.gcs.GCSCheckBlobUpdateTimeTrigger._is_blob_updated_after"
    )
    async def test_gcs_blob_update_trigger_success(self, mock_blob_updated):
        """
        Tests success case GCSCheckBlobUpdateTimeTrigger
        """
        mock_blob_updated.return_value = True, {"status": "success", "message": "success"}

        generator = self.TRIGGER.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "success", "message": "success"}) == actual

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.cloud.triggers.gcs.GCSCheckBlobUpdateTimeTrigger._is_blob_updated_after"
    )
    async def test_gcs_blob_update_trigger_pending(self, mock_blob_updated):
        """
        Test that GCSCheckBlobUpdateTimeTrigger is in loop till file isn't updated.
        """
        mock_blob_updated.return_value = False, {"status": "pending", "message": "pending"}

        task = asyncio.create_task(self.TRIGGER.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.cloud.triggers.gcs.GCSCheckBlobUpdateTimeTrigger._is_blob_updated_after"
    )
    async def test_gcs_blob_update_trigger_exception(self, mock_object_exists):
        """
        Tests the GCSCheckBlobUpdateTimeTrigger does fire if there is an exception.
        """
        mock_object_exists.side_effect = AsyncMock(side_effect=Exception("Test exception"))

        task = [i async for i in self.TRIGGER.run()]
        assert len(task) == 1
        assert TriggerEvent({"status": "error", "message": "Test exception"}) in task

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "blob_object_update_datetime, ts_object, expected_response",
        [
            (
                "2022-03-07T10:05:43.535Z",
                datetime(2022, 1, 1, 1, 1, 1),
                (True, {"status": "success", "message": "success"}),
            ),
            (
                "2022-03-07T10:05:43.535Z",
                datetime(2022, 3, 8, 1, 1, 1),
                (False, {"status": "pending", "message": "pending"}),
            ),
        ],
    )
    async def test_is_blob_updated_after(self, blob_object_update_datetime, ts_object, expected_response):
        """
        Tests to check if a particular object in Google Cloud Storage
        is found or not
        """
        hook = AsyncMock(GCSAsyncHook)
        storage = AsyncMock(Storage)
        hook.get_storage_client.return_value = storage
        bucket = AsyncMock(Bucket)
        storage.get_bucket.return_value = bucket
        bucket.get_blob.return_value.updated = blob_object_update_datetime
        trigger = GCSCheckBlobUpdateTimeTrigger(
            bucket=TEST_BUCKET,
            object_name=TEST_OBJECT,
            target_date=ts_object,
            poke_interval=TEST_POLLING_INTERVAL,
            google_cloud_conn_id=TEST_GCP_CONN_ID,
            hook_params=TEST_HOOK_PARAMS,
        )
        res = await trigger._is_blob_updated_after(hook, TEST_BUCKET, TEST_OBJECT, ts_object)
        assert res == expected_response

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "blob_object, expected_response",
        [
            (
                None,
                (
                    True,
                    {"status": "error", "message": "Object (TEST_OBJECT) not found in Bucket (TEST_BUCKET)"},
                ),
            ),
        ],
    )
    async def test_is_blob_updated_after_with_none(self, blob_object, expected_response):
        """
        Tests to check if a particular object in Google Cloud Storage
        is found or not
        """
        hook = AsyncMock(GCSAsyncHook)
        storage = AsyncMock(Storage)
        hook.get_storage_client.return_value = storage
        bucket = AsyncMock(Bucket)
        storage.get_bucket.return_value = bucket
        bucket.get_blob.return_value = blob_object

        res = await self.TRIGGER._is_blob_updated_after(hook, TEST_BUCKET, TEST_OBJECT, TEST_TS_OBJECT)
        assert res == expected_response


class TestGCSUploadSessionTrigger:
    TRIGGER = GCSUploadSessionTrigger(
        bucket=TEST_BUCKET,
        prefix=TEST_PREFIX,
        poke_interval=TEST_POLLING_INTERVAL,
        google_cloud_conn_id=TEST_GCP_CONN_ID,
        hook_params=TEST_HOOK_PARAMS,
        inactivity_period=TEST_INACTIVITY_PERIOD,
        min_objects=TEST_MIN_OBJECTS,
        previous_objects=TEST_PREVIOUS_OBJECTS,
        allow_delete=TEST_ALLOW_DELETE,
    )

    def test_gcs_upload_session_trigger_serialization(self):
        """
        Asserts that the GCSUploadSessionTrigger correctly serializes its arguments
        and classpath.
        """

        classpath, kwargs = self.TRIGGER.serialize()
        assert classpath == "airflow.providers.google.cloud.triggers.gcs.GCSUploadSessionTrigger"
        assert kwargs == {
            "bucket": TEST_BUCKET,
            "prefix": TEST_PREFIX,
            "poke_interval": TEST_POLLING_INTERVAL,
            "google_cloud_conn_id": TEST_GCP_CONN_ID,
            "hook_params": TEST_HOOK_PARAMS,
            "inactivity_period": TEST_INACTIVITY_PERIOD,
            "min_objects": TEST_MIN_OBJECTS,
            "previous_objects": TEST_PREVIOUS_OBJECTS,
            "allow_delete": TEST_ALLOW_DELETE,
        }

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSUploadSessionTrigger._list_blobs_with_prefix")
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSUploadSessionTrigger._is_bucket_updated")
    async def test_gcs_upload_session_trigger_pending(self, mock_is_bucket_updated, mock_list_blobs):
        """
        Test that GCSUploadSessionTrigger is in loop if Upload is still in progress till inactivity period.
        """
        mock_is_bucket_updated.return_value = {"status": "pending"}
        mock_list_blobs.return_value = TEST_PREVIOUS_OBJECTS

        task = asyncio.create_task(self.TRIGGER.run().__anext__())
        await asyncio.sleep(0.5)

        # TriggerEvent was not returned
        assert task.done() is False
        asyncio.get_event_loop().stop()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "is_bucket_return_value",
        [
            ({"status": "success", "message": "Successfully completed"}),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSUploadSessionTrigger._list_blobs_with_prefix")
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSUploadSessionTrigger._is_bucket_updated")
    async def test_gcs_upload_session_trigger_success(
        self, mock_is_bucket_updated, mock_list_blobs, is_bucket_return_value
    ):
        """
        Tests that the GCSUploadSessionTrigger is success case
        """
        mock_is_bucket_updated.return_value = is_bucket_return_value
        mock_list_blobs.return_value = TEST_PREVIOUS_OBJECTS

        generator = self.TRIGGER.run()
        actual = await generator.asend(None)
        assert TriggerEvent(is_bucket_return_value) == actual

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "is_bucket_return_value",
        [
            ({"status": "error", "message": "Error occurred"}),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSUploadSessionTrigger._list_blobs_with_prefix")
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSUploadSessionTrigger._is_bucket_updated")
    async def test_gcs_upload_session_trigger_error(
        self, mock_is_bucket_updated, mock_list_blobs, is_bucket_return_value
    ):
        """
        Tests that the GCSUploadSessionTrigger is error case
        """
        mock_is_bucket_updated.return_value = is_bucket_return_value
        mock_list_blobs.return_value = TEST_PREVIOUS_OBJECTS

        generator = self.TRIGGER.run()
        actual = await generator.asend(None)
        assert TriggerEvent(is_bucket_return_value) == actual

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSUploadSessionTrigger._list_blobs_with_prefix")
    async def test_gcs_upload_session_trigger_exception(self, mock_list_blobs):
        """
        Tests the GCSUploadSessionTrigger does fire if there is an exception.
        """
        mock_list_blobs.side_effect = AsyncMock(side_effect=Exception("Test exception"))

        generator = self.TRIGGER.run()
        actual = await generator.asend(None)
        assert TriggerEvent({"status": "error", "message": "Test exception"}) == actual

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "allow_delete, current_objects, response",
        [
            (True, {"a", "aa", "ab"}, {"status": "pending"}),
        ],
    )
    async def test_is_bucket_updated_pending_status(self, allow_delete, current_objects, response):
        """
        Tests to check if there is less items or more items than expected and reset the inactivity period
        along with the proper status
        """

        trigger = GCSUploadSessionTrigger(
            TEST_BUCKET,
            TEST_PREFIX,
            TEST_POLLING_INTERVAL,
            TEST_GCP_CONN_ID,
            TEST_HOOK_PARAMS,
            TEST_INACTIVITY_PERIOD,
            TEST_MIN_OBJECTS,
            TEST_PREVIOUS_OBJECTS,
            allow_delete,
        )
        res = trigger._is_bucket_updated(current_objects=current_objects)
        assert res == response

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "last_activity_time , min_objects, response",
        [
            (
                datetime.now() - timedelta(hours=0, minutes=50),
                10,
                {
                    "status": "error",
                    "message": (
                        "FAILURE: Inactivity Period passed, not enough objects found in "
                        "TEST_BUCKET/TEST_PREFIX"
                    ),
                },
            ),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.triggers.gcs.GCSUploadSessionTrigger._get_time")
    async def test_is_bucket_updated_success_failure_status(
        self, mock_time, last_activity_time, min_objects, response
    ):
        """
        Tests to check if inactivity period is finished and found min objects or not and return status
        based on that.
        """
        mock_time.return_value = (
            last_activity_time + timedelta(seconds=5) if last_activity_time else datetime.now()
        )

        trigger = GCSUploadSessionTrigger(
            TEST_BUCKET,
            TEST_PREFIX,
            TEST_POLLING_INTERVAL,
            TEST_GCP_CONN_ID,
            TEST_HOOK_PARAMS,
            TEST_INACTIVITY_PERIOD,
            min_objects,
            TEST_PREVIOUS_OBJECTS,
            TEST_ALLOW_DELETE,
        )
        trigger.last_activity_time = last_activity_time
        res = trigger._is_bucket_updated(current_objects=TEST_PREVIOUS_OBJECTS)
        assert res == response



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG that demonstrates interactions with Google Cloud Transfer.
"""

from __future__ import annotations

import os
from datetime import datetime, timedelta, timezone
from pathlib import Path

from airflow.models.dag import DAG
from airflow.providers.google.cloud.hooks.cloud_storage_transfer_service import (
    ALREADY_EXISTING_IN_SINK,
    BUCKET_NAME,
    DESCRIPTION,
    FILTER_JOB_NAMES,
    FILTER_PROJECT_ID,
    GCS_DATA_SINK,
    GCS_DATA_SOURCE,
    PROJECT_ID,
    SCHEDULE,
    SCHEDULE_END_DATE,
    SCHEDULE_START_DATE,
    START_TIME_OF_DAY,
    STATUS,
    TRANSFER_JOB,
    TRANSFER_JOB_FIELD_MASK,
    TRANSFER_OPTIONS,
    TRANSFER_SPEC,
    GcpTransferJobsStatus,
    GcpTransferOperationStatus,
)
from airflow.providers.google.cloud.operators.cloud_storage_transfer_service import (
    CloudDataTransferServiceCreateJobOperator,
    CloudDataTransferServiceDeleteJobOperator,
    CloudDataTransferServiceGetOperationOperator,
    CloudDataTransferServiceListOperationsOperator,
    CloudDataTransferServiceRunJobOperator,
    CloudDataTransferServiceUpdateJobOperator,
)
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.sensors.cloud_storage_transfer_service import (
    CloudDataTransferServiceJobStatusSensor,
)
from airflow.providers.google.cloud.transfers.local_to_gcs import LocalFilesystemToGCSOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID_TRANSFER = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")

DAG_ID = "example_gcp_transfer"

BUCKET_NAME_SRC = f"src-bucket-{DAG_ID}-{ENV_ID}".replace("_", "-")
BUCKET_NAME_DST = f"dst-bucket-{DAG_ID}-{ENV_ID}".replace("_", "-")
FILE_NAME = "transfer_service_gcp_file"
FILE_URI = f"gs://{BUCKET_NAME_SRC}/{FILE_NAME}"

CURRENT_FOLDER = Path(__file__).parent
FILE_LOCAL_PATH = str(Path(CURRENT_FOLDER) / "resources" / FILE_NAME)

# [START howto_operator_gcp_transfer_create_job_body_gcp]
gcs_to_gcs_transfer_body = {
    DESCRIPTION: "description",
    STATUS: GcpTransferJobsStatus.ENABLED,
    PROJECT_ID: PROJECT_ID_TRANSFER,
    SCHEDULE: {
        SCHEDULE_START_DATE: datetime(2015, 1, 1).date(),
        SCHEDULE_END_DATE: datetime(2030, 1, 1).date(),
        START_TIME_OF_DAY: (datetime.now(tz=timezone.utc) + timedelta(seconds=120)).time(),
    },
    TRANSFER_SPEC: {
        GCS_DATA_SOURCE: {BUCKET_NAME: BUCKET_NAME_SRC},
        GCS_DATA_SINK: {BUCKET_NAME: BUCKET_NAME_DST},
        TRANSFER_OPTIONS: {ALREADY_EXISTING_IN_SINK: True},
    },
}
# [END howto_operator_gcp_transfer_create_job_body_gcp]

# [START howto_operator_gcp_transfer_update_job_body]
update_body = {
    PROJECT_ID: PROJECT_ID_TRANSFER,
    TRANSFER_JOB: {DESCRIPTION: "description_updated"},
    TRANSFER_JOB_FIELD_MASK: "description",
}
# [END howto_operator_gcp_transfer_update_job_body]

with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "transfer", "gcp"],
) as dag:
    create_bucket_src = GCSCreateBucketOperator(
        task_id="create_bucket_src",
        bucket_name=BUCKET_NAME_SRC,
        project_id=PROJECT_ID_TRANSFER,
    )

    upload_file = LocalFilesystemToGCSOperator(
        task_id="upload_file",
        src=FILE_LOCAL_PATH,
        dst=FILE_NAME,
        bucket=BUCKET_NAME_SRC,
    )

    create_bucket_dst = GCSCreateBucketOperator(
        task_id="create_bucket_dst",
        bucket_name=BUCKET_NAME_DST,
        project_id=PROJECT_ID_TRANSFER,
    )

    create_transfer = CloudDataTransferServiceCreateJobOperator(
        task_id="create_transfer",
        body=gcs_to_gcs_transfer_body,
    )

    # [START howto_operator_gcp_transfer_update_job]
    update_transfer = CloudDataTransferServiceUpdateJobOperator(
        task_id="update_transfer",
        job_name="{{task_instance.xcom_pull('create_transfer')['name']}}",
        body=update_body,
    )
    # [END howto_operator_gcp_transfer_update_job]

    wait_for_transfer = CloudDataTransferServiceJobStatusSensor(
        task_id="wait_for_transfer",
        job_name="{{task_instance.xcom_pull('create_transfer')['name']}}",
        project_id=PROJECT_ID_TRANSFER,
        expected_statuses={GcpTransferOperationStatus.SUCCESS},
    )

    wait_for_transfer_defered = CloudDataTransferServiceJobStatusSensor(
        task_id="wait_for_transfer_defered",
        job_name="{{task_instance.xcom_pull('create_transfer')['name']}}",
        project_id=PROJECT_ID_TRANSFER,
        expected_statuses={GcpTransferOperationStatus.SUCCESS},
        deferrable=True,
    )

    # [START howto_operator_gcp_transfer_run_job]
    run_transfer = CloudDataTransferServiceRunJobOperator(
        task_id="run_transfer",
        job_name="{{task_instance.xcom_pull('create_transfer')['name']}}",
        project_id=PROJECT_ID_TRANSFER,
    )
    # [END howto_operator_gcp_transfer_run_job]

    list_operations = CloudDataTransferServiceListOperationsOperator(
        task_id="list_operations",
        request_filter={
            FILTER_PROJECT_ID: PROJECT_ID_TRANSFER,
            FILTER_JOB_NAMES: ["{{task_instance.xcom_pull('create_transfer')['name']}}"],
        },
    )

    get_operation = CloudDataTransferServiceGetOperationOperator(
        task_id="get_operation",
        operation_name="{{task_instance.xcom_pull('list_operations')[0]['name']}}",
    )

    delete_transfer = CloudDataTransferServiceDeleteJobOperator(
        task_id="delete_transfer_from_gcp_job",
        job_name="{{task_instance.xcom_pull('create_transfer')['name']}}",
        project_id=PROJECT_ID_TRANSFER,
    )

    delete_bucket_dst = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME_DST, trigger_rule=TriggerRule.ALL_DONE
    )

    delete_bucket_src = GCSDeleteBucketOperator(
        task_id="delete_bucket_src", bucket_name=BUCKET_NAME_SRC, trigger_rule=TriggerRule.ALL_DONE
    )

    (
        [create_bucket_src, create_bucket_dst]
        >> upload_file
        >> create_transfer
        >> [wait_for_transfer, wait_for_transfer_defered]
        >> update_transfer
        >> run_transfer
        >> list_operations
        >> get_operation
        >> [delete_transfer, delete_bucket_src, delete_bucket_dst]
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.dataproc_metastore import DataprocMetastoreHook
from airflow.providers.google.cloud.hooks.gcs import parse_json_from_gcs
from airflow.providers.google.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import BaseSensorOperator
else:
    from airflow.sensors.base import BaseSensorOperator  # type: ignore[no-redef]

if TYPE_CHECKING:
    from google.api_core.operation import Operation

    from airflow.utils.context import Context


class MetastoreHivePartitionSensor(BaseSensorOperator):
    """
    Waits for partitions to show up in Hive.

    This sensor uses Google Cloud SDK and passes requests via gRPC.

    :param service_id: Required. Dataproc Metastore service id.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param table: Required. Name of the partitioned table
    :param partitions: List of table partitions to wait for.
        A name of a partition should look like "ds=1", or "a=1/b=2" in case of nested partitions.
        Note that you cannot use logical or comparison operators as in HivePartitionSensor.
        If not specified then the sensor will wait for at least one partition regardless its name.
    :param gcp_conn_id: Airflow Google Cloud connection ID.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account.
    """

    template_fields: Sequence[str] = (
        "service_id",
        "region",
        "table",
        "partitions",
        "impersonation_chain",
    )

    def __init__(
        self,
        service_id: str,
        region: str,
        table: str,
        partitions: list[str] | None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.service_id = service_id
        self.region = region
        self.table = table
        self.partitions = partitions or []
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def poke(self, context: Context) -> bool:
        hook = DataprocMetastoreHook(
            gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain
        )
        operation: Operation = hook.list_hive_partitions(
            region=self.region, service_id=self.service_id, table=self.table, partition_names=self.partitions
        )
        metadata = hook.wait_for_operation(timeout=self.timeout, operation=operation)
        result_manifest_uri: str = metadata.result_manifest_uri
        self.log.info("Received result manifest URI: %s", result_manifest_uri)

        self.log.info("Extracting result manifest")
        manifest: dict = parse_json_from_gcs(
            gcp_conn_id=self.gcp_conn_id,
            file_uri=result_manifest_uri,
            impersonation_chain=self.impersonation_chain,
        )
        if not (manifest and isinstance(manifest, dict)):
            message = (
                f"Failed to extract result manifest. "
                f"Expected not empty dict, but this was received: {manifest}"
            )
            raise AirflowException(message)

        if manifest.get("status", {}).get("code") != 0:
            message = f"Request failed: {manifest.get('message')}"
            raise AirflowException(message)

        # Extract actual query results
        result_base_uri = result_manifest_uri.rsplit("/", 1)[0]
        results = (f"{result_base_uri}/{filename}" for filename in manifest.get("filenames", []))
        found_partitions = sum(
            len(
                parse_json_from_gcs(
                    gcp_conn_id=self.gcp_conn_id,
                    file_uri=uri,
                    impersonation_chain=self.impersonation_chain,
                ).get("rows", [])
            )
            for uri in results
        )

        # Return True if we got all requested partitions.
        # If no partitions were given in the request, then we expect to find at least one.
        return found_partitions >= max(1, len(set(self.partitions)))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Cloud Composer sensor."""

from __future__ import annotations

import json
from collections.abc import Iterable, Sequence
from datetime import datetime, timedelta
from functools import cached_property
from typing import TYPE_CHECKING

from dateutil import parser
from google.cloud.orchestration.airflow.service_v1.types import Environment, ExecuteAirflowCommandResponse

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.cloud_composer import CloudComposerHook
from airflow.providers.google.cloud.triggers.cloud_composer import CloudComposerDAGRunTrigger
from airflow.providers.google.common.consts import GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME
from airflow.providers.google.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import BaseSensorOperator
else:
    from airflow.sensors.base import BaseSensorOperator  # type: ignore[no-redef]
from airflow.utils.state import TaskInstanceState

if TYPE_CHECKING:
    from airflow.utils.context import Context


class CloudComposerDAGRunSensor(BaseSensorOperator):
    """
    Check if a DAG run has completed.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param environment_id: The name of the Composer environment.
    :param composer_dag_id: The ID of executable DAG.
    :param allowed_states: Iterable of allowed states, default is ``['success']``.
    :param execution_range: execution DAGs time range. Sensor checks DAGs states only for DAGs which were
        started in this time range. For yesterday, use [positive!] datetime.timedelta(days=1).
        For future, use [negative!] datetime.timedelta(days=-1). For specific time, use list of
        datetimes [datetime(2024,3,22,11,0,0), datetime(2024,3,22,12,0,0)].
        Or [datetime(2024,3,22,0,0,0)] in this case sensor will check for states from specific time in the
        past till current time execution.
        Default value datetime.timedelta(days=1).
    :param composer_dag_run_id: The Run ID of executable task. The 'execution_range' param is ignored, if both specified.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param poll_interval: Optional: Control the rate of the poll for the result of deferrable run.
    :param deferrable: Run sensor in deferrable mode.
    """

    template_fields = (
        "project_id",
        "region",
        "environment_id",
        "composer_dag_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        project_id: str,
        region: str,
        environment_id: str,
        composer_dag_id: str,
        allowed_states: Iterable[str] | None = None,
        execution_range: timedelta | list[datetime] | None = None,
        composer_dag_run_id: str | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_interval: int = 10,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.region = region
        self.environment_id = environment_id
        self.composer_dag_id = composer_dag_id
        self.allowed_states = list(allowed_states) if allowed_states else [TaskInstanceState.SUCCESS.value]
        self.execution_range = execution_range
        self.composer_dag_run_id = composer_dag_run_id
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.poll_interval = poll_interval

        if self.composer_dag_run_id and self.execution_range:
            self.log.warning(
                "The composer_dag_run_id parameter and execution_range parameter do not work together. This run will ignore execution_range parameter and count only specified composer_dag_run_id parameter."
            )

    def _get_logical_dates(self, context) -> tuple[datetime, datetime]:
        if isinstance(self.execution_range, timedelta):
            if self.execution_range < timedelta(0):
                return context["logical_date"], context["logical_date"] - self.execution_range
            return context["logical_date"] - self.execution_range, context["logical_date"]
        if isinstance(self.execution_range, list) and len(self.execution_range) > 0:
            return self.execution_range[0], self.execution_range[1] if len(
                self.execution_range
            ) > 1 else context["logical_date"]
        return context["logical_date"] - timedelta(1), context["logical_date"]

    def poke(self, context: Context) -> bool:
        start_date, end_date = self._get_logical_dates(context)

        if datetime.now(end_date.tzinfo) < end_date:
            return False

        dag_runs = self._pull_dag_runs()

        if len(dag_runs) == 0:
            self.log.info("Dag runs are empty. Sensor waits for dag runs...")
            return False

        if self.composer_dag_run_id:
            self.log.info(
                "Sensor waits for allowed states %s for specified RunID: %s",
                self.allowed_states,
                self.composer_dag_run_id,
            )
            composer_dag_run_id_status = self._check_composer_dag_run_id_states(
                dag_runs=dag_runs,
            )
            return composer_dag_run_id_status
        self.log.info("Sensor waits for allowed states: %s", self.allowed_states)
        allowed_states_status = self._check_dag_runs_states(
            dag_runs=dag_runs,
            start_date=start_date,
            end_date=end_date,
        )

        return allowed_states_status

    def _pull_dag_runs(self) -> list[dict]:
        """Pull the list of dag runs."""
        cmd_parameters = (
            ["-d", self.composer_dag_id, "-o", "json"]
            if self._composer_airflow_version < 3
            else [self.composer_dag_id, "-o", "json"]
        )
        dag_runs_cmd = self.hook.execute_airflow_command(
            project_id=self.project_id,
            region=self.region,
            environment_id=self.environment_id,
            command="dags",
            subcommand="list-runs",
            parameters=cmd_parameters,
        )
        cmd_result = self.hook.wait_command_execution_result(
            project_id=self.project_id,
            region=self.region,
            environment_id=self.environment_id,
            execution_cmd_info=ExecuteAirflowCommandResponse.to_dict(dag_runs_cmd),
        )
        dag_runs = json.loads(cmd_result["output"][0]["content"])
        return dag_runs

    def _check_dag_runs_states(
        self,
        dag_runs: list[dict],
        start_date: datetime,
        end_date: datetime,
    ) -> bool:
        for dag_run in dag_runs:
            if (
                start_date.timestamp()
                < parser.parse(
                    dag_run["execution_date" if self._composer_airflow_version < 3 else "logical_date"]
                ).timestamp()
                < end_date.timestamp()
            ) and dag_run["state"] not in self.allowed_states:
                return False
        return True

    def _get_composer_airflow_version(self) -> int:
        """Return Composer Airflow version."""
        environment_obj = self.hook.get_environment(
            project_id=self.project_id,
            region=self.region,
            environment_id=self.environment_id,
        )
        environment_config = Environment.to_dict(environment_obj)
        image_version = environment_config["config"]["software_config"]["image_version"]
        return int(image_version.split("airflow-")[1].split(".")[0])

    def _check_composer_dag_run_id_states(self, dag_runs: list[dict]) -> bool:
        for dag_run in dag_runs:
            if dag_run["run_id"] == self.composer_dag_run_id and dag_run["state"] in self.allowed_states:
                return True
        return False

    def execute(self, context: Context) -> None:
        self._composer_airflow_version = self._get_composer_airflow_version()
        if self.deferrable:
            start_date, end_date = self._get_logical_dates(context)
            self.defer(
                timeout=self.timeout,
                trigger=CloudComposerDAGRunTrigger(
                    project_id=self.project_id,
                    region=self.region,
                    environment_id=self.environment_id,
                    composer_dag_id=self.composer_dag_id,
                    composer_dag_run_id=self.composer_dag_run_id,
                    start_date=start_date,
                    end_date=end_date,
                    allowed_states=self.allowed_states,
                    gcp_conn_id=self.gcp_conn_id,
                    impersonation_chain=self.impersonation_chain,
                    poll_interval=self.poll_interval,
                    composer_airflow_version=self._composer_airflow_version,
                ),
                method_name=GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME,
            )
        super().execute(context)

    def execute_complete(self, context: Context, event: dict):
        if event and event["status"] == "error":
            raise AirflowException(event["message"])
        self.log.info("DAG %s has executed successfully.", self.composer_dag_id)

    @cached_property
    def hook(self) -> CloudComposerHook:
        return CloudComposerHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that uses Google Cloud Run Service Operators.
"""

from __future__ import annotations

import os
from datetime import datetime

from google.cloud.run_v2 import Service
from google.cloud.run_v2.types import k8s_min

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.cloud_run import (
    CloudRunCreateServiceOperator,
    CloudRunDeleteServiceOperator,
)

PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")


# [START howto_operator_cloud_run_service_creation]
def _create_service():
    service = Service()
    container = k8s_min.Container()
    container.image = "us-docker.pkg.dev/cloudrun/container/placeholder:latest"
    service.template.containers.append(container)
    return service


# [END howto_operator_cloud_run_service_creation]


with DAG(
    "cloud_run_service",
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["cloud-run", "service", "example"],
) as dag:
    # [START howto_operator_cloud_run_create_service]
    create_cloud_run_service = CloudRunCreateServiceOperator(
        task_id="create-cloud-run-service",
        project_id=PROJECT_ID,
        region="us-central1",
        service=_create_service(),
        service_name="cloudrun-system-test-service",
    )
    # [END howto_operator_cloud_run_create_service]

    # [START howto_operator_cloud_run_delete_service]
    delete_cloud_run_service = CloudRunDeleteServiceOperator(
        task_id="delete-cloud-run-service",
        project_id=PROJECT_ID,
        region="us-central1",
        service_name="cloudrun-system-test-service",
        dag=dag,
    )
    # [END howto_operator_cloud_run_delete_service]

    (
        # TEST SETUP
        # TEST BODY
        create_cloud_run_service
        # TEST TEARDOWN
        >> delete_cloud_run_service
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains operator for uploading local file(s) to GCS."""

from __future__ import annotations

import os
from collections.abc import Sequence
from glob import glob
from typing import TYPE_CHECKING

from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.utils.context import Context


class LocalFilesystemToGCSOperator(BaseOperator):
    """
    Uploads a file or list of files to Google Cloud Storage; optionally can compress the file for upload; optionally can upload the data in multiple chunks.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:LocalFilesystemToGCSOperator`

    :param src: Path to the local file, or list of local files. Path can be either absolute
        (e.g. /path/to/file.ext) or relative (e.g. ../../foo/*/*.csv). (templated)
    :param dst: Destination path within the specified bucket on GCS (e.g. /path/to/file.ext).
        If multiple files are being uploaded, specify object prefix with trailing backslash
        (e.g. /path/to/directory/) (templated)
    :param bucket: The bucket to upload to. (templated)
    :param gcp_conn_id: (Optional) The connection ID used to connect to Google Cloud.
    :param mime_type: The mime-type string
    :param gzip: Allows for file to be compressed and uploaded as gzip
    :param chunk_size: Blob chunk size in bytes. This must be a multiple of 262144 bytes (256 KiB)
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "src",
        "dst",
        "bucket",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        src: str | list[str],
        dst: str,
        bucket: str,
        gcp_conn_id: str = "google_cloud_default",
        mime_type: str = "application/octet-stream",
        gzip: bool = False,
        chunk_size: int | None = None,
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.src = src
        self.dst = dst
        self.bucket = bucket
        self.gcp_conn_id = gcp_conn_id
        self.mime_type = mime_type
        self.gzip = gzip
        self.chunk_size = chunk_size
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        """Upload a file or list of files to Google Cloud Storage."""
        hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        filepaths = self.src if isinstance(self.src, list) else glob(self.src)
        if not filepaths:
            raise FileNotFoundError(self.src)
        if os.path.basename(self.dst):  # path to a file
            if len(filepaths) > 1:  # multiple file upload
                raise ValueError(
                    "'dst' parameter references filepath. Please specify "
                    "directory (with trailing backslash) to upload multiple "
                    "files. e.g. /path/to/directory/"
                )
            object_paths = [self.dst]
        else:  # directory is provided
            object_paths = [os.path.join(self.dst, os.path.basename(filepath)) for filepath in filepaths]

        for filepath, object_path in zip(filepaths, object_paths):
            hook.upload(
                bucket_name=self.bucket,
                object_name=object_path,
                mime_type=self.mime_type,
                filename=filepath,
                gzip=self.gzip,
                chunk_size=self.chunk_size,
            )

    def get_openlineage_facets_on_start(self):
        from airflow.providers.common.compat.openlineage.facet import (
            Dataset,
            Identifier,
            SymlinksDatasetFacet,
        )
        from airflow.providers.google.cloud.openlineage.utils import WILDCARD, extract_ds_name_from_gcs_path
        from airflow.providers.openlineage.extractors import OperatorLineage

        source_facets = {}
        if isinstance(self.src, str):  # Single path provided, possibly relative or with wildcard
            original_src = f"{self.src}"
            absolute_src = os.path.abspath(self.src)
            resolved_src = extract_ds_name_from_gcs_path(absolute_src)
            if original_src.startswith("/") and not resolved_src.startswith("/"):
                resolved_src = "/" + resolved_src
            source_objects = [resolved_src]

            if WILDCARD in original_src or absolute_src != resolved_src:
                # We attach a symlink with unmodified path.
                source_facets = {
                    "symlink": SymlinksDatasetFacet(
                        identifiers=[Identifier(namespace="file", name=original_src, type="file")]
                    ),
                }
        else:
            source_objects = self.src

        dest_object = self.dst if os.path.basename(self.dst) else extract_ds_name_from_gcs_path(self.dst)

        return OperatorLineage(
            inputs=[Dataset(namespace="file", name=src, facets=source_facets) for src in source_objects],
            outputs=[Dataset(namespace=f"gs://{self.bucket}", name=dest_object)],
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Sequence
from typing import TYPE_CHECKING, Any, SupportsAbs

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientResponseError
from asgiref.sync import sync_to_async

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.bigquery import BigQueryAsyncHook, BigQueryTableAsyncHook
from airflow.providers.google.version_compat import AIRFLOW_V_3_0_PLUS
from airflow.triggers.base import BaseTrigger, TriggerEvent
from airflow.utils.state import TaskInstanceState

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session

if not AIRFLOW_V_3_0_PLUS:
    from airflow.models.taskinstance import TaskInstance
    from airflow.utils.session import provide_session


class BigQueryInsertJobTrigger(BaseTrigger):
    """
    BigQueryInsertJobTrigger run on the trigger worker to perform insert operation.

    :param conn_id: Reference to google cloud connection id
    :param job_id:  The ID of the job. It will be suffixed with hash of job configuration
    :param project_id: Google Cloud Project where the job is running
    :param location: The dataset location.
    :param dataset_id: The dataset ID of the requested table. (templated)
    :param table_id: The table ID of the requested table. (templated)
    :param poll_interval: polling period in seconds to check for the status. (templated)
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account. (templated)
    """

    def __init__(
        self,
        conn_id: str,
        job_id: str | None,
        project_id: str,
        location: str | None,
        dataset_id: str | None = None,
        table_id: str | None = None,
        poll_interval: float = 4.0,
        impersonation_chain: str | Sequence[str] | None = None,
        cancel_on_kill: bool = True,
    ):
        super().__init__()
        self.log.info("Using the connection  %s .", conn_id)
        self.conn_id = conn_id
        self.job_id = job_id
        self._job_conn = None
        self.dataset_id = dataset_id
        self.project_id = project_id
        self.location = location
        self.table_id = table_id
        self.poll_interval = poll_interval
        self.impersonation_chain = impersonation_chain
        self.cancel_on_kill = cancel_on_kill

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize BigQueryInsertJobTrigger arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.bigquery.BigQueryInsertJobTrigger",
            {
                "conn_id": self.conn_id,
                "job_id": self.job_id,
                "dataset_id": self.dataset_id,
                "project_id": self.project_id,
                "location": self.location,
                "table_id": self.table_id,
                "poll_interval": self.poll_interval,
                "impersonation_chain": self.impersonation_chain,
                "cancel_on_kill": self.cancel_on_kill,
            },
        )

    if not AIRFLOW_V_3_0_PLUS:

        @provide_session
        def get_task_instance(self, session: Session) -> TaskInstance:
            query = session.query(TaskInstance).filter(
                TaskInstance.dag_id == self.task_instance.dag_id,
                TaskInstance.task_id == self.task_instance.task_id,
                TaskInstance.run_id == self.task_instance.run_id,
                TaskInstance.map_index == self.task_instance.map_index,
            )
            task_instance = query.one_or_none()
            if task_instance is None:
                raise AirflowException(
                    "TaskInstance with dag_id: %s, task_id: %s, run_id: %s and map_index: %s is not found",
                    self.task_instance.dag_id,
                    self.task_instance.task_id,
                    self.task_instance.run_id,
                    self.task_instance.map_index,
                )
            return task_instance

    async def get_task_state(self):
        from airflow.sdk.execution_time.task_runner import RuntimeTaskInstance

        task_states_response = await sync_to_async(RuntimeTaskInstance.get_task_states)(
            dag_id=self.task_instance.dag_id,
            task_ids=[self.task_instance.task_id],
            run_ids=[self.task_instance.run_id],
            map_index=self.task_instance.map_index,
        )
        try:
            task_state = task_states_response[self.task_instance.run_id][self.task_instance.task_id]
        except Exception:
            raise AirflowException(
                "TaskInstance with dag_id: %s, task_id: %s, run_id: %s and map_index: %s is not found",
                self.task_instance.dag_id,
                self.task_instance.task_id,
                self.task_instance.run_id,
                self.task_instance.map_index,
            )
        return task_state

    async def safe_to_cancel(self) -> bool:
        """
        Whether it is safe to cancel the external job which is being executed by this trigger.

        This is to avoid the case that `asyncio.CancelledError` is called because the trigger itself is stopped.
        Because in those cases, we should NOT cancel the external job.
        """
        if AIRFLOW_V_3_0_PLUS:
            task_state = await self.get_task_state()
        else:
            # Database query is needed to get the latest state of the task instance.
            task_instance = self.get_task_instance()  # type: ignore[call-arg]
            task_state = task_instance.state
        return task_state != TaskInstanceState.DEFERRED

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Get current job execution status and yields a TriggerEvent."""
        hook = self._get_async_hook()
        try:
            while True:
                job_status = await hook.get_job_status(
                    job_id=self.job_id, project_id=self.project_id, location=self.location
                )
                if job_status["status"] == "success":
                    yield TriggerEvent(
                        {
                            "job_id": self.job_id,
                            "status": job_status["status"],
                            "message": job_status["message"],
                        }
                    )
                    return
                elif job_status["status"] == "error":
                    yield TriggerEvent(job_status)
                    return
                else:
                    self.log.info(
                        "Bigquery job status is %s. Sleeping for %s seconds.",
                        job_status["status"],
                        self.poll_interval,
                    )
                    await asyncio.sleep(self.poll_interval)
        except asyncio.CancelledError:
            if self.job_id and self.cancel_on_kill and await self.safe_to_cancel():
                self.log.info(
                    "The job is safe to cancel the as airflow TaskInstance is not in deferred state."
                )
                self.log.info(
                    "Cancelling job. Project ID: %s, Location: %s, Job ID: %s",
                    self.project_id,
                    self.location,
                    self.job_id,
                )
                await hook.cancel_job(job_id=self.job_id, project_id=self.project_id, location=self.location)
            else:
                self.log.info(
                    "Trigger may have shutdown. Skipping to cancel job because the airflow "
                    "task is not cancelled yet: Project ID: %s, Location:%s, Job ID:%s",
                    self.project_id,
                    self.location,
                    self.job_id,
                )
        except Exception as e:
            self.log.exception("Exception occurred while checking for query completion")
            yield TriggerEvent({"status": "error", "message": str(e)})

    def _get_async_hook(self) -> BigQueryAsyncHook:
        return BigQueryAsyncHook(gcp_conn_id=self.conn_id, impersonation_chain=self.impersonation_chain)


class BigQueryCheckTrigger(BigQueryInsertJobTrigger):
    """BigQueryCheckTrigger run on the trigger worker."""

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize BigQueryCheckTrigger arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.bigquery.BigQueryCheckTrigger",
            {
                "conn_id": self.conn_id,
                "job_id": self.job_id,
                "dataset_id": self.dataset_id,
                "project_id": self.project_id,
                "location": self.location,
                "table_id": self.table_id,
                "poll_interval": self.poll_interval,
                "impersonation_chain": self.impersonation_chain,
                "cancel_on_kill": self.cancel_on_kill,
            },
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Get current job execution status and yields a TriggerEvent."""
        hook = self._get_async_hook()
        try:
            while True:
                # Poll for job execution status
                job_status = await hook.get_job_status(job_id=self.job_id, project_id=self.project_id)
                if job_status["status"] == "success":
                    query_results = await hook.get_job_output(job_id=self.job_id, project_id=self.project_id)
                    records = hook.get_records(query_results)

                    # If empty list, then no records are available
                    if not records:
                        yield TriggerEvent(
                            {
                                "status": job_status["status"],
                                "message": job_status["message"],
                                "records": None,
                            }
                        )
                    else:
                        # Extract only first record from the query results
                        first_record = records.pop(0)
                        yield TriggerEvent(
                            {
                                "status": job_status["status"],
                                "message": job_status["message"],
                                "records": first_record,
                            }
                        )
                    return
                elif job_status["status"] == "error":
                    yield TriggerEvent({"status": "error", "message": job_status["message"]})
                    return
                else:
                    self.log.info(
                        "Bigquery job status is %s. Sleeping for %s seconds.",
                        job_status["status"],
                        self.poll_interval,
                    )
                    await asyncio.sleep(self.poll_interval)
        except Exception as e:
            self.log.exception("Exception occurred while checking for query completion")
            yield TriggerEvent({"status": "error", "message": str(e)})


class BigQueryGetDataTrigger(BigQueryInsertJobTrigger):
    """
    BigQueryGetDataTrigger run on the trigger worker, inherits from BigQueryInsertJobTrigger class.

    :param as_dict: if True returns the result as a list of dictionaries, otherwise as list of lists
        (default: False).
    """

    def __init__(self, as_dict: bool = False, selected_fields: str | None = None, **kwargs):
        super().__init__(**kwargs)
        self.as_dict = as_dict
        self.selected_fields = selected_fields

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize BigQueryInsertJobTrigger arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.bigquery.BigQueryGetDataTrigger",
            {
                "conn_id": self.conn_id,
                "job_id": self.job_id,
                "dataset_id": self.dataset_id,
                "project_id": self.project_id,
                "location": self.location,
                "table_id": self.table_id,
                "poll_interval": self.poll_interval,
                "impersonation_chain": self.impersonation_chain,
                "as_dict": self.as_dict,
                "selected_fields": self.selected_fields,
            },
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Get current job execution status and yields a TriggerEvent with response data."""
        hook = self._get_async_hook()
        try:
            while True:
                # Poll for job execution status
                job_status = await hook.get_job_status(job_id=self.job_id, project_id=self.project_id)
                if job_status["status"] == "success":
                    query_results = await hook.get_job_output(job_id=self.job_id, project_id=self.project_id)
                    records = hook.get_records(
                        query_results=query_results,
                        as_dict=self.as_dict,
                        selected_fields=self.selected_fields,
                    )
                    self.log.debug("Response from hook: %s", job_status["status"])
                    yield TriggerEvent(
                        {
                            "status": "success",
                            "message": job_status["message"],
                            "records": records,
                        }
                    )
                    return
                elif job_status["status"] == "error":
                    yield TriggerEvent(job_status)
                    return
                else:
                    self.log.info(
                        "Bigquery job status is %s. Sleeping for %s seconds.",
                        job_status["status"],
                        self.poll_interval,
                    )
                    await asyncio.sleep(self.poll_interval)
        except Exception as e:
            self.log.exception("Exception occurred while checking for query completion")
            yield TriggerEvent({"status": "error", "message": str(e)})


class BigQueryIntervalCheckTrigger(BigQueryInsertJobTrigger):
    """
    BigQueryIntervalCheckTrigger run on the trigger worker, inherits from BigQueryInsertJobTrigger class.

    :param conn_id: Reference to google cloud connection id
    :param first_job_id:  The ID of the job 1 performed
    :param second_job_id:  The ID of the job 2 performed
    :param project_id: Google Cloud Project where the job is running
    :param dataset_id: The dataset ID of the requested table. (templated)
    :param table: table name
    :param metrics_thresholds: dictionary of ratios indexed by metrics
    :param location: The dataset location.
    :param date_filter_column: column name. (templated)
    :param days_back: number of days between ds and the ds we want to check against. (templated)
    :param ratio_formula: ration formula. (templated)
    :param ignore_zero: boolean value to consider zero or not. (templated)
    :param table_id: The table ID of the requested table. (templated)
    :param poll_interval: polling period in seconds to check for the status. (templated)
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account. (templated)
    """

    def __init__(
        self,
        conn_id: str,
        first_job_id: str,
        second_job_id: str,
        project_id: str,
        table: str,
        metrics_thresholds: dict[str, int],
        location: str | None = None,
        date_filter_column: str | None = "ds",
        days_back: SupportsAbs[int] = -7,
        ratio_formula: str = "max_over_min",
        ignore_zero: bool = True,
        dataset_id: str | None = None,
        table_id: str | None = None,
        poll_interval: float = 4.0,
        impersonation_chain: str | Sequence[str] | None = None,
    ):
        super().__init__(
            conn_id=conn_id,
            job_id=first_job_id,
            project_id=project_id,
            location=location,
            dataset_id=dataset_id,
            table_id=table_id,
            poll_interval=poll_interval,
            impersonation_chain=impersonation_chain,
        )
        self.conn_id = conn_id
        self.first_job_id = first_job_id
        self.second_job_id = second_job_id
        self.project_id = project_id
        self.table = table
        self.metrics_thresholds = metrics_thresholds
        self.date_filter_column = date_filter_column
        self.days_back = days_back
        self.ratio_formula = ratio_formula
        self.ignore_zero = ignore_zero

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize BigQueryCheckTrigger arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.bigquery.BigQueryIntervalCheckTrigger",
            {
                "conn_id": self.conn_id,
                "first_job_id": self.first_job_id,
                "second_job_id": self.second_job_id,
                "project_id": self.project_id,
                "table": self.table,
                "metrics_thresholds": self.metrics_thresholds,
                "location": self.location,
                "date_filter_column": self.date_filter_column,
                "days_back": self.days_back,
                "ratio_formula": self.ratio_formula,
                "ignore_zero": self.ignore_zero,
                "impersonation_chain": self.impersonation_chain,
            },
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Get current job execution status and yields a TriggerEvent."""
        hook = self._get_async_hook()
        try:
            while True:
                first_job_response_from_hook = await hook.get_job_status(
                    job_id=self.first_job_id, project_id=self.project_id
                )
                second_job_response_from_hook = await hook.get_job_status(
                    job_id=self.second_job_id, project_id=self.project_id
                )

                if (
                    first_job_response_from_hook["status"] == "success"
                    and second_job_response_from_hook["status"] == "success"
                ):
                    first_query_results = await hook.get_job_output(
                        job_id=self.first_job_id, project_id=self.project_id
                    )

                    second_query_results = await hook.get_job_output(
                        job_id=self.second_job_id, project_id=self.project_id
                    )

                    first_records = hook.get_records(first_query_results)

                    second_records = hook.get_records(second_query_results)

                    # If empty list, then no records are available
                    if not first_records:
                        first_job_row: str | None = None
                    else:
                        # Extract only first record from the query results
                        first_job_row = first_records.pop(0)

                    # If empty list, then no records are available
                    if not second_records:
                        second_job_row: str | None = None
                    else:
                        # Extract only first record from the query results
                        second_job_row = second_records.pop(0)

                    hook.interval_check(
                        first_job_row,
                        second_job_row,
                        self.metrics_thresholds,
                        self.ignore_zero,
                        self.ratio_formula,
                    )

                    yield TriggerEvent(
                        {
                            "status": "success",
                            "message": "Job completed",
                            "first_row_data": first_job_row,
                            "second_row_data": second_job_row,
                        }
                    )
                    return
                elif (
                    first_job_response_from_hook["status"] == "pending"
                    or second_job_response_from_hook["status"] == "pending"
                ):
                    self.log.info("Query is still running...")
                    self.log.info("Sleeping for %s seconds.", self.poll_interval)
                    await asyncio.sleep(self.poll_interval)
                else:
                    yield TriggerEvent(
                        {"status": "error", "message": second_job_response_from_hook["message"], "data": None}
                    )
                    return

        except Exception as e:
            self.log.exception("Exception occurred while checking for query completion")
            yield TriggerEvent({"status": "error", "message": str(e)})


class BigQueryValueCheckTrigger(BigQueryInsertJobTrigger):
    """
    BigQueryValueCheckTrigger run on the trigger worker, inherits from BigQueryInsertJobTrigger class.

    :param conn_id: Reference to google cloud connection id
    :param sql: the sql to be executed
    :param pass_value: pass value
    :param job_id: The ID of the job
    :param project_id: Google Cloud Project where the job is running
    :param tolerance: certain metrics for tolerance. (templated)
    :param dataset_id: The dataset ID of the requested table. (templated)
    :param table_id: The table ID of the requested table. (templated)
    :param location: The dataset location
    :param poll_interval: polling period in seconds to check for the status. (templated)
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    def __init__(
        self,
        conn_id: str,
        sql: str,
        pass_value: int | float | str,
        job_id: str | None,
        project_id: str,
        tolerance: Any = None,
        dataset_id: str | None = None,
        table_id: str | None = None,
        location: str | None = None,
        poll_interval: float = 4.0,
        impersonation_chain: str | Sequence[str] | None = None,
    ):
        super().__init__(
            conn_id=conn_id,
            job_id=job_id,
            project_id=project_id,
            dataset_id=dataset_id,
            table_id=table_id,
            location=location,
            poll_interval=poll_interval,
            impersonation_chain=impersonation_chain,
        )
        self.sql = sql
        self.pass_value = pass_value
        self.tolerance = tolerance

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize BigQueryValueCheckTrigger arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.bigquery.BigQueryValueCheckTrigger",
            {
                "conn_id": self.conn_id,
                "pass_value": self.pass_value,
                "job_id": self.job_id,
                "dataset_id": self.dataset_id,
                "project_id": self.project_id,
                "sql": self.sql,
                "table_id": self.table_id,
                "tolerance": self.tolerance,
                "location": self.location,
                "poll_interval": self.poll_interval,
                "impersonation_chain": self.impersonation_chain,
            },
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Get current job execution status and yields a TriggerEvent."""
        hook = self._get_async_hook()
        try:
            while True:
                # Poll for job execution status
                response_from_hook = await hook.get_job_status(job_id=self.job_id, project_id=self.project_id)
                if response_from_hook["status"] == "success":
                    query_results = await hook.get_job_output(job_id=self.job_id, project_id=self.project_id)
                    records = hook.get_records(query_results)
                    _records = records.pop(0) if records else None
                    hook.value_check(self.sql, self.pass_value, _records, self.tolerance)
                    yield TriggerEvent({"status": "success", "message": "Job completed", "records": _records})
                    return
                elif response_from_hook["status"] == "pending":
                    self.log.info("Query is still running...")
                    self.log.info("Sleeping for %s seconds.", self.poll_interval)
                    await asyncio.sleep(self.poll_interval)
                else:
                    yield TriggerEvent(
                        {"status": "error", "message": response_from_hook["message"], "records": None}
                    )
                    return
        except Exception as e:
            self.log.exception("Exception occurred while checking for query completion")
            yield TriggerEvent({"status": "error", "message": str(e)})


class BigQueryTableExistenceTrigger(BaseTrigger):
    """
    Initialize the BigQuery Table Existence Trigger with needed parameters.

    :param project_id: Google Cloud Project where the job is running
    :param dataset_id: The dataset ID of the requested table.
    :param table_id: The table ID of the requested table.
    :param gcp_conn_id: Reference to google cloud connection id
    :param hook_params: params for hook
    :param poll_interval: polling period in seconds to check for the status
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account. (templated)
    """

    def __init__(
        self,
        project_id: str,
        dataset_id: str,
        table_id: str,
        gcp_conn_id: str,
        hook_params: dict[str, Any],
        poll_interval: float = 4.0,
        impersonation_chain: str | Sequence[str] | None = None,
    ):
        self.dataset_id = dataset_id
        self.project_id = project_id
        self.table_id = table_id
        self.gcp_conn_id: str = gcp_conn_id
        self.poll_interval = poll_interval
        self.hook_params = hook_params
        self.impersonation_chain = impersonation_chain

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize BigQueryTableExistenceTrigger arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.bigquery.BigQueryTableExistenceTrigger",
            {
                "dataset_id": self.dataset_id,
                "project_id": self.project_id,
                "table_id": self.table_id,
                "gcp_conn_id": self.gcp_conn_id,
                "poll_interval": self.poll_interval,
                "hook_params": self.hook_params,
                "impersonation_chain": self.impersonation_chain,
            },
        )

    def _get_async_hook(self) -> BigQueryTableAsyncHook:
        return BigQueryTableAsyncHook(
            gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Will run until the table exists in the Google Big Query."""
        try:
            while True:
                hook = self._get_async_hook()
                response = await self._table_exists(
                    hook=hook, dataset=self.dataset_id, table_id=self.table_id, project_id=self.project_id
                )
                if response:
                    yield TriggerEvent({"status": "success", "message": "success"})
                    return
                await asyncio.sleep(self.poll_interval)
        except Exception as e:
            self.log.exception("Exception occurred while checking for Table existence")
            yield TriggerEvent({"status": "error", "message": str(e)})

    async def _table_exists(
        self, hook: BigQueryTableAsyncHook, dataset: str, table_id: str, project_id: str
    ) -> bool:
        """
        Create session, make call to BigQueryTableAsyncHook, and check for the table in Google Big Query.

        :param hook: BigQueryTableAsyncHook Hook class
        :param dataset:  The name of the dataset in which to look for the table storage bucket.
        :param table_id: The name of the table to check the existence of.
        :param project_id: The Google cloud project in which to look for the table.
            The connection supplied to the hook must provide
            access to the specified project.
        """
        async with ClientSession() as session:
            try:
                client = await hook.get_table_client(
                    dataset=dataset, table_id=table_id, project_id=project_id, session=session
                )
                response = await client.get()
                return bool(response)
            except ClientResponseError as err:
                if err.status == 404:
                    return False
                raise err


class BigQueryTablePartitionExistenceTrigger(BigQueryTableExistenceTrigger):
    """
    Initialize the BigQuery Table Partition Existence Trigger with needed parameters.

    :param partition_id: The name of the partition to check the existence of.
    :param project_id: Google Cloud Project where the job is running
    :param dataset_id: The dataset ID of the requested table.
    :param table_id: The table ID of the requested table.
    :param gcp_conn_id: Reference to google cloud connection id
    :param hook_params: params for hook
    :param poll_interval: polling period in seconds to check for the status
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account. (templated)
    """

    def __init__(self, partition_id: str, **kwargs):
        super().__init__(**kwargs)
        self.partition_id = partition_id

    def serialize(self) -> tuple[str, dict[str, Any]]:
        """Serialize BigQueryTablePartitionExistenceTrigger arguments and classpath."""
        return (
            "airflow.providers.google.cloud.triggers.bigquery.BigQueryTablePartitionExistenceTrigger",
            {
                "partition_id": self.partition_id,
                "dataset_id": self.dataset_id,
                "project_id": self.project_id,
                "table_id": self.table_id,
                "gcp_conn_id": self.gcp_conn_id,
                "poll_interval": self.poll_interval,
                "impersonation_chain": self.impersonation_chain,
                "hook_params": self.hook_params,
            },
        )

    async def run(self) -> AsyncIterator[TriggerEvent]:
        """Will run until the table exists in the Google Big Query."""
        hook = BigQueryAsyncHook(gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain)
        job_id = None
        while True:
            if job_id is not None:
                job_status = await hook.get_job_status(job_id=job_id, project_id=self.project_id)
                if job_status["status"] == "success":
                    is_partition = await self._partition_exists(
                        hook=hook, job_id=job_id, project_id=self.project_id
                    )
                    if is_partition:
                        yield TriggerEvent(
                            {
                                "status": "success",
                                "message": f"Partition: {self.partition_id} in table: {self.table_id}",
                            }
                        )
                        return
                    job_id = None
                elif job_status["status"] == "error":
                    yield TriggerEvent(job_status)
                    return
                self.log.info("Sleeping for %s seconds.", self.poll_interval)
                await asyncio.sleep(self.poll_interval)

            else:
                job_id = await hook.create_job_for_partition_get(
                    self.dataset_id, table_id=self.table_id, project_id=self.project_id
                )
                self.log.info("Sleeping for %s seconds.", self.poll_interval)
                await asyncio.sleep(self.poll_interval)

    async def _partition_exists(self, hook: BigQueryAsyncHook, job_id: str | None, project_id: str):
        query_results = await hook.get_job_output(job_id=job_id, project_id=project_id)
        records = hook.get_records(query_results)
        if records:
            records = [row[0] for row in records]
            return self.partition_id in records



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from google.cloud.batch import ListJobsRequest
from google.cloud.batch_v1 import CreateJobRequest, Job, JobStatus

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.cloud_batch import CloudBatchAsyncHook, CloudBatchHook

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

pytestmark = pytest.mark.db_test


class TestCloudBathHook:
    def dummy_get_credentials(self):
        pass

    @pytest.fixture
    def cloud_batch_hook(self):
        cloud_batch_hook = CloudBatchHook()
        cloud_batch_hook.get_credentials = self.dummy_get_credentials
        return cloud_batch_hook

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_submit(self, mock_batch_service_client, cloud_batch_hook):
        job = Job()
        job_name = "jobname"
        project_id = "test_project_id"
        region = "us-central1"

        cloud_batch_hook.submit_batch_job(
            job_name=job_name, job=Job.to_dict(job), region=region, project_id=project_id
        )

        create_request = CreateJobRequest()
        create_request.job = job
        create_request.job_id = job_name
        create_request.parent = f"projects/{project_id}/locations/{region}"

        cloud_batch_hook._client.create_job.assert_called_with(create_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_get_job(self, mock_batch_service_client, cloud_batch_hook):
        cloud_batch_hook.get_job("job1")
        cloud_batch_hook._client.get_job.assert_called_once_with(name="job1")

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_delete_job(self, mock_batch_service_client, cloud_batch_hook):
        job_name = "job1"
        region = "us-east1"
        project_id = "test_project_id"
        cloud_batch_hook.delete_job(job_name=job_name, region=region, project_id=project_id)
        cloud_batch_hook._client.delete_job.assert_called_once_with(
            name=f"projects/{project_id}/locations/{region}/jobs/{job_name}"
        )

    @pytest.mark.parametrize("state", [JobStatus.State.SUCCEEDED])
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_wait_job_succeeded(self, mock_batch_service_client, state, cloud_batch_hook):
        mock_job = self._mock_job_with_status(state)
        mock_batch_service_client.return_value.get_job.return_value = mock_job
        actual_job = cloud_batch_hook.wait_for_job("job1")
        assert actual_job == mock_job

    @pytest.mark.parametrize("state", [JobStatus.State.FAILED, JobStatus.State.DELETION_IN_PROGRESS])
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_wait_job_does_not_succeed(self, mock_batch_service_client, state, cloud_batch_hook):
        mock_job = self._mock_job_with_status(state)
        mock_batch_service_client.return_value.get_job.return_value = mock_job
        with pytest.raises(AirflowException):
            cloud_batch_hook.wait_for_job("job1")

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_wait_job_timeout(self, mock_batch_service_client, cloud_batch_hook):
        mock_job = self._mock_job_with_status(JobStatus.State.RUNNING)
        mock_batch_service_client.return_value.get_job.return_value = mock_job

        exception_caught = False
        try:
            cloud_batch_hook.wait_for_job("job1", polling_period_seconds=0.01, timeout=0.02)
        except AirflowException:
            exception_caught = True

        assert exception_caught

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_list_jobs(self, mock_batch_service_client, cloud_batch_hook):
        number_of_jobs = 3
        region = "us-central1"
        project_id = "test_project_id"
        filter = "filter_description"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_batch_hook.list_jobs(region=region, project_id=project_id, filter=filter)

        for i in range(number_of_jobs):
            assert jobs_list[i].name == f"name{i}"

        expected_list_jobs_request: ListJobsRequest = ListJobsRequest(
            parent=f"projects/{project_id}/locations/{region}", filter=filter
        )
        mock_batch_service_client.return_value.list_jobs.assert_called_once_with(
            request=expected_list_jobs_request
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_list_jobs_with_limit(self, mock_batch_service_client, cloud_batch_hook):
        number_of_jobs = 3
        limit = 2
        region = "us-central1"
        project_id = "test_project_id"
        filter = "filter_description"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_batch_hook.list_jobs(
            region=region, project_id=project_id, filter=filter, limit=limit
        )

        assert len(jobs_list) == limit
        for i in range(limit):
            assert jobs_list[i].name == f"name{i}"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_list_jobs_with_limit_zero(self, mock_batch_service_client, cloud_batch_hook):
        number_of_jobs = 3
        limit = 0
        region = "us-central1"
        project_id = "test_project_id"
        filter = "filter_description"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_batch_hook.list_jobs(
            region=region, project_id=project_id, filter=filter, limit=limit
        )

        assert len(jobs_list) == 0

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_list_jobs_with_limit_greater_then_range(self, mock_batch_service_client, cloud_batch_hook):
        number_of_jobs = 3
        limit = 5
        region = "us-central1"
        project_id = "test_project_id"
        filter = "filter_description"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_batch_hook.list_jobs(
            region=region, project_id=project_id, filter=filter, limit=limit
        )

        assert len(jobs_list) == number_of_jobs
        for i in range(number_of_jobs):
            assert jobs_list[i].name == f"name{i}"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_list_jobs_with_limit_less_than_zero(self, mock_batch_service_client, cloud_batch_hook):
        number_of_jobs = 3
        limit = -1
        region = "us-central1"
        project_id = "test_project_id"
        filter = "filter_description"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        with pytest.raises(expected_exception=AirflowException):
            cloud_batch_hook.list_jobs(region=region, project_id=project_id, filter=filter, limit=limit)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_list_tasks_with_limit(self, mock_batch_service_client, cloud_batch_hook):
        number_of_tasks = 3
        limit = 2
        region = "us-central1"
        project_id = "test_project_id"
        filter = "filter_description"
        job_name = "test_job"

        page = self._mock_pager(number_of_tasks)
        mock_batch_service_client.return_value.list_tasks.return_value = page

        tasks_list = cloud_batch_hook.list_tasks(
            region=region, project_id=project_id, job_name=job_name, filter=filter, limit=limit
        )

        assert len(tasks_list) == limit
        for i in range(limit):
            assert tasks_list[i].name == f"name{i}"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_list_tasks_with_limit_greater_then_range(self, mock_batch_service_client, cloud_batch_hook):
        number_of_tasks = 3
        limit = 5
        region = "us-central1"
        project_id = "test_project_id"
        filter = "filter_description"
        job_name = "test_job"

        page = self._mock_pager(number_of_tasks)
        mock_batch_service_client.return_value.list_tasks.return_value = page

        tasks_list = cloud_batch_hook.list_tasks(
            region=region, project_id=project_id, filter=filter, job_name=job_name, limit=limit
        )

        assert len(tasks_list) == number_of_tasks
        for i in range(number_of_tasks):
            assert tasks_list[i].name == f"name{i}"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceClient")
    def test_list_tasks_with_limit_less_than_zero(self, mock_batch_service_client, cloud_batch_hook):
        number_of_tasks = 3
        limit = -1
        region = "us-central1"
        project_id = "test_project_id"
        filter = "filter_description"
        job_name = "test_job"

        page = self._mock_pager(number_of_tasks)
        mock_batch_service_client.return_value.list_tasks.return_value = page

        with pytest.raises(expected_exception=AirflowException):
            cloud_batch_hook.list_tasks(
                region=region, project_id=project_id, job_name=job_name, filter=filter, limit=limit
            )

    def _mock_job_with_status(self, status: JobStatus.State):
        job = mock.MagicMock()
        job.status.state = status
        return job

    def _mock_pager(self, number_of_jobs):
        mock_pager = []
        for i in range(number_of_jobs):
            mock_pager.append(Job(name=f"name{i}"))

        return mock_pager


class TestCloudBatchAsyncHook:
    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_batch.BatchServiceAsyncClient")
    async def test_get_job(self, mock_client):
        expected_job = {"name": "somename"}

        async def _get_job(name):
            return expected_job

        job_name = "jobname"
        mock_client.return_value = mock.MagicMock()
        mock_client.return_value.get_job = _get_job

        hook = CloudBatchAsyncHook()
        hook.get_credentials = self._dummy_get_credentials

        returned_operation = await hook.get_batch_job(job_name=job_name)

        assert returned_operation == expected_job

    def _dummy_get_credentials(self):
        pass



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from asyncio import Future
from copy import deepcopy
from unittest import mock
from unittest.mock import AsyncMock

import pytest
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.bigquery_datatransfer_v1.types import TransferConfig

from airflow.providers.google.cloud.hooks.bigquery_dts import (
    AsyncBiqQueryDataTransferServiceHook,
    BiqQueryDataTransferServiceHook,
)

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_no_default_project_id

CREDENTIALS = "test-creds"
PROJECT_ID = "id"

PARAMS = {
    "field_delimiter": ",",
    "max_bad_records": 0,
    "skip_leading_rows": 1,
    "data_path_template": "bucket",
    "destination_table_name_template": "name",
    "file_format": "CSV",
}

TRANSFER_CONFIG = TransferConfig(
    destination_dataset_id="dataset",
    display_name="GCS Test Config",
    data_source_id="google_cloud_storage",
    params=PARAMS,
)

TRANSFER_CONFIG_ID = "id1234"
RUN_ID = "id1234"


class TestBigQueryDataTransferHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.cloud.hooks.bigquery_dts.GoogleBaseHook.__init__",
            new=mock_base_gcp_hook_no_default_project_id,
        ):
            self.hook = BiqQueryDataTransferServiceHook()
            self.hook.get_credentials = mock.MagicMock(return_value=CREDENTIALS)

    def test_disable_auto_scheduling(self):
        expected = deepcopy(TRANSFER_CONFIG)
        expected.schedule_options.disable_auto_scheduling = True
        assert expected == self.hook._disable_auto_scheduling(TRANSFER_CONFIG)

    @mock.patch(
        "airflow.providers.google.cloud.hooks.bigquery_dts.DataTransferServiceClient.create_transfer_config"
    )
    def test_create_transfer_config(self, service_mock):
        self.hook.create_transfer_config(transfer_config=TRANSFER_CONFIG, project_id=PROJECT_ID)
        parent = f"projects/{PROJECT_ID}"
        expected_config = deepcopy(TRANSFER_CONFIG)
        expected_config.schedule_options.disable_auto_scheduling = True
        service_mock.assert_called_once_with(
            request=dict(parent=parent, transfer_config=expected_config, authorization_code=None),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.bigquery_dts.DataTransferServiceClient.delete_transfer_config"
    )
    def test_delete_transfer_config(self, service_mock):
        self.hook.delete_transfer_config(transfer_config_id=TRANSFER_CONFIG_ID, project_id=PROJECT_ID)

        name = f"projects/{PROJECT_ID}/transferConfigs/{TRANSFER_CONFIG_ID}"
        service_mock.assert_called_once_with(
            request=dict(name=name), metadata=(), retry=DEFAULT, timeout=None
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.bigquery_dts."
        "DataTransferServiceClient.start_manual_transfer_runs"
    )
    def test_start_manual_transfer_runs(self, service_mock):
        self.hook.start_manual_transfer_runs(transfer_config_id=TRANSFER_CONFIG_ID, project_id=PROJECT_ID)

        parent = f"projects/{PROJECT_ID}/transferConfigs/{TRANSFER_CONFIG_ID}"
        service_mock.assert_called_once_with(
            request=dict(parent=parent, requested_time_range=None, requested_run_time=None),
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )

    @mock.patch(
        "airflow.providers.google.cloud.hooks.bigquery_dts.DataTransferServiceClient.get_transfer_run"
    )
    def test_get_transfer_run(self, service_mock):
        self.hook.get_transfer_run(
            run_id=RUN_ID, transfer_config_id=TRANSFER_CONFIG_ID, project_id=PROJECT_ID
        )

        name = f"projects/{PROJECT_ID}/transferConfigs/{TRANSFER_CONFIG_ID}/runs/{RUN_ID}"
        service_mock.assert_called_once_with(
            request=dict(name=name), metadata=(), retry=DEFAULT, timeout=None
        )


class TestAsyncBiqQueryDataTransferServiceHook:
    HOOK_MODULE_PATH = "airflow.providers.google.cloud.hooks.bigquery_dts"

    @pytest.fixture
    def mock_client(self):
        with mock.patch(
            f"{self.HOOK_MODULE_PATH}.AsyncBiqQueryDataTransferServiceHook._get_conn",
            new_callable=AsyncMock,
        ) as mock_client:
            transfer_result = Future()
            transfer_result.set_result(mock.MagicMock())

            mock_client.return_value.get_transfer_run = mock.MagicMock(return_value=transfer_result)
            yield mock_client

    @pytest.fixture
    def hook(self):
        return AsyncBiqQueryDataTransferServiceHook()

    @pytest.mark.db_test
    @pytest.mark.asyncio
    async def test_get_transfer_run(self, mock_client, hook):
        await hook.get_transfer_run(
            run_id=RUN_ID,
            config_id=TRANSFER_CONFIG_ID,
            project_id=PROJECT_ID,
        )

        name = f"projects/{PROJECT_ID}/transferConfigs/{TRANSFER_CONFIG_ID}/runs/{RUN_ID}"
        mock_client.return_value.get_transfer_run.assert_called_once_with(
            name=name,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG for testing Google Dataflow to create pipelines.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.dataflow import (
    DataflowCreatePipelineOperator,
    DataflowDeletePipelineOperator,
    DataflowRunPipelineOperator,
)
from airflow.providers.google.cloud.operators.gcs import (
    GCSCreateBucketOperator,
    GCSDeleteBucketOperator,
    GCSSynchronizeBucketsOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

DAG_ID = "dataflow_pipeline"
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
GCP_PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
GCP_LOCATION = "us-central1"

PIPELINE_NAME = f"{DAG_ID}-{ENV_ID}".replace("_", "-")
PIPELINE_JOB_NAME = f"{DAG_ID}-{ENV_ID}-job".replace("_", "-")
PIPELINE_TYPE = "PIPELINE_TYPE_BATCH"

RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"
BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}".replace("-", "_")

FILE_NAME = "kinglear.txt"
TEMPLATE_FILE = "word-count.json"

TEMP_LOCATION = f"gs://{BUCKET_NAME}/temp"

GCS_PATH = f"gs://{BUCKET_NAME}/dataflow/{TEMPLATE_FILE}"
INPUT_FILE = f"gs://{BUCKET_NAME}/dataflow/{FILE_NAME}"
OUTPUT = f"gs://{BUCKET_NAME}/results/hello"

with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "dataflow", "pipelines"],
) as dag:
    create_bucket = GCSCreateBucketOperator(task_id="create_bucket", bucket_name=BUCKET_NAME)

    move_files_to_bucket = GCSSynchronizeBucketsOperator(
        task_id="move_files_to_bucket",
        source_bucket=RESOURCE_DATA_BUCKET,
        source_object="dataflow/pipelines",
        destination_bucket=BUCKET_NAME,
        destination_object="dataflow",
        recursive=True,
    )

    # [START howto_operator_create_dataflow_pipeline]
    create_pipeline = DataflowCreatePipelineOperator(
        task_id="create_pipeline",
        project_id=GCP_PROJECT_ID,
        location=GCP_LOCATION,
        body={
            "name": f"projects/{GCP_PROJECT_ID}/locations/{GCP_LOCATION}/pipelines/{PIPELINE_NAME}",
            "type": PIPELINE_TYPE,
            "workload": {
                "dataflowFlexTemplateRequest": {
                    "launchParameter": {
                        "containerSpecGcsPath": GCS_PATH,
                        "jobName": PIPELINE_JOB_NAME,
                        "environment": {"tempLocation": TEMP_LOCATION},
                        "parameters": {
                            "inputFile": INPUT_FILE,
                            "output": OUTPUT,
                        },
                    },
                    "projectId": GCP_PROJECT_ID,
                    "location": GCP_LOCATION,
                }
            },
        },
    )
    # [END howto_operator_create_dataflow_pipeline]

    # [START howto_operator_run_dataflow_pipeline]
    run_pipeline = DataflowRunPipelineOperator(
        task_id="run_pipeline",
        pipeline_name=PIPELINE_NAME,
        project_id=GCP_PROJECT_ID,
    )
    # [END howto_operator_run_dataflow_pipeline]

    # [START howto_operator_delete_dataflow_pipeline]
    delete_pipeline = DataflowDeletePipelineOperator(
        task_id="delete_pipeline",
        pipeline_name=PIPELINE_NAME,
        project_id=GCP_PROJECT_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_delete_dataflow_pipeline]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    (
        # TEST SETUP
        create_bucket
        >> move_files_to_bucket
        # TEST BODY
        >> create_pipeline
        >> run_pipeline
        # TEST TEARDOWN
        >> delete_pipeline
        >> delete_bucket
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.google.cloud.hooks.gcs import _parse_gcs_url
from airflow.providers.google.version_compat import AIRFLOW_V_3_0_PLUS

if TYPE_CHECKING:
    from urllib.parse import SplitResult

    from airflow.providers.common.compat.assets import Asset
    from airflow.providers.common.compat.openlineage.facet import Dataset as OpenLineageDataset
else:
    try:
        from airflow.providers.common.compat.assets import Asset
    except ImportError:
        if AIRFLOW_V_3_0_PLUS:
            from airflow.sdk.definitions.asset import Asset
        else:
            # dataset is renamed to asset since Airflow 3.0
            from airflow.datasets import Dataset as Asset


def create_asset(*, bucket: str, key: str, extra: dict | None = None) -> Asset:
    return Asset(uri=f"gs://{bucket}/{key}", extra=extra)


def sanitize_uri(uri: SplitResult) -> SplitResult:
    if not uri.netloc:
        raise ValueError("URI format gs:// must contain a bucket name")
    return uri


def convert_asset_to_openlineage(asset: Asset, lineage_context) -> OpenLineageDataset:
    """Translate Asset with valid AIP-60 uri to OpenLineage with assistance from the hook."""
    from airflow.providers.common.compat.openlineage.facet import Dataset as OpenLineageDataset

    bucket, key = _parse_gcs_url(asset.uri)
    return OpenLineageDataset(namespace=f"gs://{bucket}", name=key if key else "/")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.cloud.translate_v3.types import TranslateTextResponse

from airflow.providers.google.cloud.hooks.translate import CloudTranslateHook, TranslateHook
from airflow.providers.google.common.consts import CLIENT_INFO

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

PROJECT_ID_TEST = "project-id"


class TestCloudTranslateHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.cloud.hooks.translate.CloudTranslateHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = CloudTranslateHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.translate.CloudTranslateHook.get_credentials")
    @mock.patch("airflow.providers.google.cloud.hooks.translate.Client")
    def test_translate_client_creation(self, mock_client, mock_get_creds):
        result = self.hook.get_conn()
        mock_client.assert_called_once_with(credentials=mock_get_creds.return_value, client_info=CLIENT_INFO)
        assert mock_client.return_value == result
        assert self.hook._client == result

    @mock.patch("airflow.providers.google.cloud.hooks.translate.CloudTranslateHook.get_conn")
    def test_translate_called(self, get_conn):
        # Given
        translate_method = get_conn.return_value.translate
        translate_method.return_value = {
            "translatedText": "Yellowing self Gęśle",
            "detectedSourceLanguage": "pl",
            "model": "base",
            "input": "zażółć gęślą jaźń",
        }
        # When
        result = self.hook.translate(
            values=["zażółć gęślą jaźń"],
            target_language="en",
            format_="text",
            source_language=None,
            model="base",
        )
        # Then
        assert result == {
            "translatedText": "Yellowing self Gęśle",
            "detectedSourceLanguage": "pl",
            "model": "base",
            "input": "zażółć gęślą jaźń",
        }
        translate_method.assert_called_once_with(
            values=["zażółć gęślą jaźń"],
            target_language="en",
            format_="text",
            source_language=None,
            model="base",
        )


class TestTranslateHook:
    def setup_method(self):
        with mock.patch(
            "airflow.providers.google.cloud.hooks.translate.TranslateHook.__init__",
            new=mock_base_gcp_hook_default_project_id,
        ):
            self.hook = TranslateHook(gcp_conn_id="test")

    @mock.patch("airflow.providers.google.cloud.hooks.translate.TranslateHook.get_credentials")
    @mock.patch("airflow.providers.google.cloud.hooks.translate.TranslationServiceClient")
    def test_translate_client_creation(self, mock_client, mock_get_creds):
        result = self.hook.get_client()
        mock_client.assert_called_once_with(credentials=mock_get_creds.return_value, client_info=CLIENT_INFO)
        assert mock_client.return_value == result
        assert self.hook._client == result

    @mock.patch("airflow.providers.google.cloud.hooks.translate.TranslateHook.get_client")
    def test_translate_text_method(self, get_client):
        translation_result_data = {
            "translations": [
                {"translated_text": "Hello World!", "model": "", "detected_language_code": ""},
                {
                    "translated_text": "Can you get me a cup of coffee, please?",
                    "model": "",
                    "detected_language_code": "",
                },
            ],
            "glossary_translations": [],
        }
        data_to_translate = ["Ciao mondo!", "Mi puoi prendere una tazza di caffè, per favore?"]
        translate_client = get_client.return_value
        translate_text_client_method = translate_client.translate_text
        translate_text_client_method.return_value = TranslateTextResponse(translation_result_data)

        input_translation_args = dict(
            project_id=PROJECT_ID_TEST,
            contents=data_to_translate,
            source_language_code="it",
            target_language_code="en",
            mime_type="text/plain",
            location="global",
            glossary_config=None,
            transliteration_config=None,
            model=None,
            labels=None,
            metadata=(),
            timeout=30,
            retry=None,
        )
        result = self.hook.translate_text(**input_translation_args)
        assert result == translation_result_data

        expected_call_args = {
            "request": {
                "parent": f"projects/{PROJECT_ID_TEST}/locations/global",
                "contents": data_to_translate,
                "source_language_code": "it",
                "target_language_code": "en",
                "mime_type": "text/plain",
                "glossary_config": None,
                "transliteration_config": None,
                "model": None,
                "labels": None,
            },
            "retry": None,
            "metadata": (),
            "timeout": 30,
        }
        translate_text_client_method.assert_called_once_with(**expected_call_args)

    @mock.patch("airflow.providers.google.cloud.hooks.translate.TranslateHook.get_client")
    def test_batch_translate_text_method(self, get_client):
        sample_method_result = "batch_translate_api_call_result_obj"
        translate_client = get_client.return_value
        translate_text_client_method = translate_client.batch_translate_text
        translate_text_client_method.return_value = sample_method_result
        BATCH_TRANSLATE_INPUT = {
            "gcs_source": {"input_uri": "input_source_uri"},
            "mime_type": "text/plain",
        }
        GCS_OUTPUT_DST = {"gcs_destination": {"output_uri_prefix": "translate_output_uri_prefix"}}
        LOCATION = "us-central1"
        input_translation_args = dict(
            project_id=PROJECT_ID_TEST,
            source_language_code="de",
            target_language_codes=["en", "uk"],
            location=LOCATION,
            input_configs=[BATCH_TRANSLATE_INPUT],
            output_config=GCS_OUTPUT_DST,
            glossaries=None,
            models=None,
            labels=None,
            metadata=(),
            timeout=30,
            retry=None,
        )
        result = self.hook.batch_translate_text(**input_translation_args)
        expected_call_args = {
            "request": {
                "parent": f"projects/{PROJECT_ID_TEST}/locations/{LOCATION}",
                "source_language_code": "de",
                "target_language_codes": ["en", "uk"],
                "input_configs": [BATCH_TRANSLATE_INPUT],
                "output_config": GCS_OUTPUT_DST,
                "glossaries": None,
                "models": None,
                "labels": None,
            },
            "retry": None,
            "metadata": (),
            "timeout": 30,
        }
        translate_text_client_method.assert_called_once_with(**expected_call_args)
        assert result == sample_method_result
